import os, pprint
from bioblend import galaxy

# import docopt
try:
    from docopt import docopt
except ImportError:
    sys.exit("""Error: docopt is required by 'galaxy-api-tools'.
https://pypi.python.org/pypi/docopt""")

def upload(options):
    """
usage:
  galaxy-api-tools upload (--library <name=l_name_or_id=l_id> | --history <name=h_name_or_id=h_id>) [options] [--] <file_path>...
    [--add-file <ext=type>]... [--workflow-input <input=dataset_id=src>]... [--workflow-param <tool=param=value>]...

options:
  -h, --help                            Show this screen.

  -l <l_inf>, --library <l_inf>         Upload files to this data library, with
                                        <l_inf> in the form of <name=l_name> or
                                        <id=l_id>.
  -s <h_inf>, --history <h_inf>         Upload files to this history, with
                                        <h_inf> in the form of <name=h_name>
                                        or <id=h_id>.
  -f <f_name>, --folder <f_name>        Folder in data library where to place
                                        uploaded files [default: /].
  -b <f_id>, --base-folder <f_id>       Existing folder in data library to use
                                        as root folder. If not specified data
                                        library root folder is used.
  -d <desc>, --description <desc>       Description will be used when creating
                                        the folder.
  -l, --link                            Link data into Galaxy. By default data
                                        is copied instead. Can only be used when
                                        uploading to data libraries using upload
                                        methods 'from_server' or
                                        'galaxy_filesystem'.
  -m <method>, --upload-method <method> Upload method to use. Can be any of
                                        these: 'local_path', 'from_server',
                                        'from_url' or 'galaxy_filesystem'
                                        [default: local_path].
  -g <dbkey>, --genome <dbkey>          Assign <dbkey> to the imported data
                                        [default: ?].
  -t <type>, --file-type <type>         File type [default: auto].
  -a <add_inf>, --add-file <add_inf>    Process additional file per <file_path>
                                        with <add_inf> in the form of
                                        <extension=type[=label]>. A matching
                                        file with provided extension is expected
                                        in the same path. Type 'auto' is
                                        allowed. Labels are required if a
                                        workflow needs to be executed,
                                        otherwise should not be provided. Main
                                        file will be used as input for any label
                                        with no file association.
  -w <wf_inf>, --workflow <wf_inf>      Run this workflow on uploaded datasets,
                                        with <wf_inf> in the form of
                                        <name=wf_name> or <id=wf_id>.
  -i <i_inf>, --workflow-input <i_inf>  Set extra workflow inputs, with <i_inf>
                                        in the form of <input=dataset_id=src>.
  -p <p_inf>, --workflow-param <p_inf>  Set workflow tools parameter, with
                                        <p_inf> in the form of
                                        <tool=param=value>.
  -k <api_key>, --api-key <api_key>     User API key. Can be also set by using
                                        environment variable 'GALAXY_API_KEY'.
  -u <galaxy_url>, --url <galaxy_url>   Galxy server URL. Can be also set by
                                        using environment variable 'GALAXY_URL'.

When provinding inputs labels, you can use '.' as a wildcard. There can only be
one input using this wildcard. This input will be mapped to all workflow inputs
with no other input argument provided.

examples:
    galaxy-api-tools upload --library id=f597429621d6eb2b 1.fna
    galaxy-api-tools upload --history name="API Imports" 1.fna
    galaxy-api-tools upload --library name="API Imports" \
        --upload-method galaxy_filesystem --genome hg19 --folder "Run 20130918" \
        --workflow name="Combine and Split" --add-file .qual=qual454=QUAL \
        --workflow-input BARCODE_INDEX=1cd8e2f6b131e891=ldda \
        --workflow-param ngs-tools_split_by_barcode=barcodes=RL001,RL002 \
        /local/seq_data/R_2013_09_18/1.fna /local/seq_data/R_2013_09_18/2.fna
    """

    options_upload = docopt(upload.__doc__, argv=options)

    library = dict([options_upload['--library'].split('=')]) if options_upload['--library'] else None
    history = dict([options_upload['--history'].split('=')]) if options_upload['--history'] else None
    folder = {'name': options_upload['--folder']}
    base_folder = {'id': options_upload['--base-folder']}
    description = options_upload['--description']
    method = options_upload['--upload-method']
    link_data_only = options_upload['--link']
    dbkey = options_upload['--genome']
    file_type = options_upload['--file-type']
    files = list(set(options_upload['<file_path>']))
    workflow_inputs = options_upload['--workflow-input']
    workflow_params = _parse_params(options_upload['--workflow-param']) if options_upload['--workflow-param'] else None

    # create Bioblend's GalaxyInstance
    gi = _create_galaxy_instance(url=options_upload['--url'], key=options_upload['--api-key'])

    # get workflow
    workflow = dict([options_upload['--workflow'].split('=')]) if options_upload['--workflow'] else None
    if workflow:
        workflow = _get_workflow(gi, workflow)

    # process additional files
    add_files = []
    labels = {}
    for add_f in options_upload['--add-file']:
        try:
            if workflow:
                ext, g_type, label = add_f.split('=')
            else:
                ext, g_type = add_f.split('=')
                label = None
        except ValueError:
            raise ValueError("There was a problem parsing additional file: {0}. Please make sure you are using the proper argument syntax, --add-file extension=type[=label].".format(add_f))
        add_files.append({'ext': ext.lstrip('.'), 'type': g_type, 'label': label})
        # if workflow provided check label mapping is correct
        if workflow:
            if ext in labels:
                raise ValueError("The relation label-to-extension can only be 1-to-1. Error: extension '{0}' is been linked to more than one label.".format(ext))
            if label in labels.values():
                raise ValueError("The relation label-to-extension can only be 1-to-1. Error: label '{0}' is been linked to more than one extension.".format(label))
            labels[ext] = label

    # find library and folder if provided
    if library:
        try:
            ll = []
            if 'id' in library:
                ll = gi.libraries.get_libraries(library_id=library['id'])
                if not ll:
                    raise ValueError("There was a problem pulling information for library with id '{0}'. Please check library id.".format(library['id']))
                if base_folder['id'] and not _get_folders(gi, library['id'], folder_id=base_folder['id']):
                    raise ValueError("There was a problem pulling information for base folder with id '{0}'. Please check base folder id.".format(base_folder['id']))
            elif 'name' in library:
                ll = gi.libraries.get_libraries(name=library['name'])
                if len(ll) > 1:
                    ll_info = ['id: {0}, name: {1}'.format(l['id'], l['name']) for l in ll]
                    raise ValueError("There are more than one library with name '{0}'. Please use a more specific info like library id. Matching libraries: {1}".format(library['name'], ll_info))
                if ll:
                    library['id'] = ll[0]['id']
                    if base_folder['id'] and not _get_folders(gi, library_id['id'], folder_id=base_folder['id']):
                        raise ValueError("There was a problem pulling information for base folder with id '{0}'. Please check base folder id.".format(base_folder['id']))
            else:
                raise ValueError("There was a problem parsing library 'id' or 'name'. Please make sure you are using the proper argument syntax, --library id=library_id or --library name=library_name.")
        except:
            raise

    # check and generate uploading data information
    upload_files = []
    for f in files:
        try:
            file_info = {
                    'file_location' : f,
                    'file_type' : file_type,
                    }
            if workflow:
                file_info['label'] = '.'
            if add_files:
                file_info['add_files'] = []
                for add_f in add_files:
                    add_file_info = {
                            'file_location' : "{0}.{1}".format(os.path.splitext(f)[0], add_f['ext']),
                            'file_type' : add_f['type'],
                            'label': add_f['label'],
                            }
                    file_info['add_files'].append(add_file_info)
            upload_files.append(file_info)
        except KeyError as e:
            print "Cannot generate uploading data information. Error: KeyError: {0}.".format(e)
            raise ValueError("Please make sure you provided all arguments correctly. Hint: if you are providing any label mapping at all, you need to make sure you provide label mappings for all extensions.")

    # start doing things
    if history:
        history = _get_history(gi, history)
        for f in upload_files:
            _do_upload(gi, f, history_id=history['id'], workflow=workflow, workflow_inputs=workflow_inputs, workflow_params=workflow_params, label_ext_map=label_ext_map, dbkey=dbkey)
    elif library:
        # create data library if it doesn't exist
        if not ll:
            ll = gi.libraries.create_library(library['name'])
            library['id'] = ll['id']
        # create folder if provided
        folder['id'] = None
        if folder['name'] is not '/':
            folder_json = gi.libraries.create_folder(library['id'], folder['name'], description=description, base_folder_id=base_folder['id'])
            folder['id'] = folder_json[0]['id']
        # do upload and execute workflow if provided
        for f in upload_files:
            _do_upload(gi, f, library_id=library['id'], workflow=workflow, workflow_inputs=workflow_inputs, workflow_params=workflow_params, dbkey=dbkey, method=method, folder=folder, link_data_only=link_data_only)

def run_workflow(options):
    """
usage:
  galaxy-api-tools run-workflow [options] [--param <tool=param=value>] ... 
    [--] <name=wf_name_or_id=wf_id> <input=dataset_id=src> ...

    
options:
  -h, --help                            Show this screen.
  -s <h_inf>, --history <h_inf>         Store result files in this history, with
                                        <h_inf> in the form of <name=h_name>
                                        or <id=h_id>.
  -p <p_inf>, --param <p_inf>           Set workflow tool parameter.
  -i, --import-inputs                   Import workflow inputs into history.
  -k <api_key>, --api-key <api_key>     User API key. Can be also set by using
                                        environment variable 'GALAXY_API_KEY'.
  -u <galaxy_url>, --url <galaxy_url>   Galxy server URL. Can be also set by
                                        using environment variable 'GALAXY_URL'.

When provinding inputs labels, you can use '.' as a wildcard. There can only be
one input using this wildcard. This input will be mapped to all workflow inputs
with no other input argument provided.

examples:
    galaxy-api-tools run-workflow id=f2db41e1fa331b3e \
        FASTA=cbbbf59e8f08c98c=ldda QUAL=964b37715ec9bd22=ldda \
        BARCODE_INDEX=1cd8e2f6b131e891=ldda
    galaxy-api-tools run-workflow \
        --param ngs-tools_split_by_barcode=barcodes=RL001,RL002 \
        name="Combine and Split" \
        FASTA=cbbbf59e8f08c98c=ldda QUAL=964b37715ec9bd22=ldda \
        BARCODE_INDEX=1cd8e2f6b131e891=ldda
    """

    options_run_workflow = docopt(run_workflow.__doc__, argv=options)

    history = dict([options_run_workflow['--history'].split('=')]) if options_run_workflow['--history'] else {}
    params = _parse_params(options_run_workflow['--param']) if options_run_workflow['--param'] else None
    import_inputs = options_run_workflow['--import-inputs']

    # create Bioblend's GalaxyInstance
    gi = _create_galaxy_instance(url=options_run_workflow['--url'], key=options_run_workflow['--api-key'])

    # get workflow
    workflow = dict([options_run_workflow['<name=wf_name_or_id=wf_id>'].split('=')])
    workflow = _get_workflow(gi, workflow)

    #parse inputs
    inputs = options_run_workflow['<input=dataset_id=src>']
    inputs = _parse_inputs(inputs, workflow)

    if not history:
        history = {'name': "{0}".format(workflow['name'])}
    history = _get_history(gi, history)

    run_json = gi.workflows.run_workflow(workflow['id'], inputs, params=params, history_id=history['id'], import_inputs_to_history=import_inputs)
    pp = pprint.PrettyPrinter(indent=4)
    pp.pprint(run_json)

def get_workflows(options):
    """
usage:
  galaxy-api-tools get-workflows [options] [--] [<name=wf_name_or_id=wf_id>]

options:
  -h, --help                            Show this screen.

  -s, --show                            Show workflows detailed information.
  -k <api_key>, --api-key <api_key>     User API key. Can be also set by using
                                        environment variable 'GALAXY_API_KEY'.
  -u <galaxy_url>, --url <galaxy_url>   Galxy server URL. Can be also set by
                                        using environment variable 'GALAXY_URL'.
examples:
    galaxy-api-tools get-workflows
    galaxy-api-tools get-workflows name="Combine and Split"
    galaxy-api-tools get-workflows id=f2db41e1fa331b3e
    """

    options_get_workflows = docopt(get_workflows.__doc__, argv=options)
    show = options_get_workflows['--show']

    # create Bioblend's GalaxyInstance
    gi = _create_galaxy_instance(url=options_get_workflows['--url'], key=options_get_workflows['--api-key'])

    # get workflows
    workflow = dict([options_get_workflows['<name=wf_name_or_id=wf_id>'].split('=')]) if options_get_workflows['<name=wf_name_or_id=wf_id>'] else {'id': None, 'name': None}
    workflows = _get_workflows(gi, workflow)

    pp = pprint.PrettyPrinter(indent=4)
    if show:
        for wf in workflows:
            pp.pprint(gi.workflows.show_workflow(wf['id']))
    else:
        pp.pprint(workflows)

def show_workflow(options):
    """
usage:
  galaxy-api-tools show-workflow [options] <name=wf_name_or_id=wf_id>

options:
  -h, --help                            Show this screen.

  -k <api_key>, --api-key <api_key>     User API key. Can be also set by using
                                        environment variable 'GALAXY_API_KEY'.
  -u <galaxy_url>, --url <galaxy_url>   Galxy server URL. Can be also set by
                                        using environment variable 'GALAXY_URL'.
examples:
    galaxy-api-tools show-workflow name="Combine and Split"
    galaxy-api-tools show-workflow id=f2db41e1fa331b3e
    """

    options_show_workflow = docopt(show_workflow.__doc__, argv=options)

    # create Bioblend's GalaxyInstance
    gi = _create_galaxy_instance(url=options_show_workflow['--url'], key=options_show_workflow['--api-key'])

    # get workflow
    workflow = dict([options_show_workflow['<name=wf_name_or_id=wf_id>'].split('=')])
    workflow = _get_workflow(gi, workflow)

    pp = pprint.PrettyPrinter(indent=4)
    pp.pprint(workflow)

def _create_galaxy_instance(url, key):
    if not key:
        key = os.getenv('GALAXY_API_KEY')
    if not key:
        raise ValueError("Please provide a Galaxy user API key by using option '--api-key' or by setting enviroment variable 'GALAXY_API_KEY'.")

    if not url:
        url = os.getenv('GALAXY_URL')
    if not url:
        raise ValueError("Please provide the Galaxy server URL by using option '--url' or by setting enviroment variable 'GALAXY_URL'.")
    url = url.rstrip('/')

    # create Bioblend's GalaxyInstance
    return galaxy.GalaxyInstance(url=url, key=key)

def _get_folders(gi, library_id, folder_id=None, name=None, deleted=False):
    """
    Get all the folders or filter specific one(s) via the provided ``name``
    or ``folder_id`` in data library ``library_id``. Provide only one
    argument: ``name`` or ``folder_id``, but not both.

    If ``name`` is set and multiple names match the given name, all the
    folders matching the argument will be returned.

    If ``deleted`` is set to ``True``, return folders that have been deleted.

    Return a list of JSON formatted dicts each containing basic information
    about a folder.
    """
    library_contents = gi.libraries.show_library(library_id, contents=True)
    folders = []
    filtered_folders = []
    for content in library_contents:
        if content['type'] == 'folder':
            folders.append(content)
            if (name is not None or folder_id is not None) and (name == content['name'] or folder_id == content['id']):
                filtered_folders.append(content)
            if folder_id is not None and filtered_folders:
                break
    if name is not None or folder_id is not None:
        folders = filtered_folders
    return folders

def _get_workflows(gi, workflow):

    if 'name' in workflow:
        name = workflow['name']
    else:
        name = None
    if 'id' in workflow:
        workflow_id = workflow['id']
    else:
        workflow_id = None

    workflows = gi.workflows.get_workflows()
    if name is not None or workflow_id is not None:
        filtered_wfs = []
        for workflow in workflows:
            if name == workflow['name'] or workflow_id == workflow['id']:
                filtered_wfs.append(workflow)
                # Workflows ID's are unique so break now that the wf was found
                if workflow_id is not None:
                    break
        workflows = filtered_wfs
    return workflows

def _get_workflow(gi, workflow, deleted=False):

    if 'name' not in workflow and 'id' not in workflow:
        raise ValueError("There was a problem parsing workflow 'id' or 'name'. Please make sure you are using the proper argument syntax, id=workflow_id or name=workflow_name.")

    workflows = _get_workflows(gi, workflow)

    if len(workflows) > 1:
        raise ValueError("More than one workflow match the supplied information. Workflows matching: \n{0}".format(workflows))
    elif not workflows:
        raise ValueError("No workflow was found matching the supplied information: {0}.".format(workflow))

    return gi.workflows.show_workflow(workflows[0]['id'])

def _get_history(gi, history, deleted=False):
    if 'id' in history:
        history = gi.histories.get_histories(history_id=history['id'])
        if not history:
            raise ValueError("There was a problem pulling information for history with id '{0}'. Please check history id.".format(history['id']))
    elif 'name' in history:
        history = gi.histories.create_history(name=history['name'])
    else:
        raise ValueError("There was a problem parsing history information. Please make sure you are using the proper argument syntax, id=history_id or name=history_name.")

    return history

def _upload_to_library(gi, file_info, method, library_id, folder_id=None, dbkey='?', link_data_only=None):
    if method == 'local_path':
        upload_json = gi.libraries.upload_file_from_local_path(library_id, file_info['file_location'], folder_id=folder_id, file_type=file_info['file_type'], dbkey=dbkey)
    elif method == 'from_server':
        upload_json = gi.libraries.upload_file_from_server(library_id, file_info['file_location'], folder_id=folder_id, file_type=file_info['file_type'], dbkey=dbkey, link_data_only=link_data_only)
    elif method == 'from_url':
        upload_json = gi.libraries.upload_file_from_url(library_id, file_info['file_location'], folder_id=folder_id, file_type=file_info['file_type'], dbkey=dbkey)
    elif method == 'galaxy_filesystem':
        upload_json = gi.libraries.upload_from_galaxy_filesystem(library_id, file_info['file_location'], folder_id=folder_id, file_type=file_info['file_type'], dbkey=dbkey, link_data_only=link_data_only)
    else:
        raise ValueError("Unknown upload method.")
    return upload_json
 
def _do_upload(gi, file_info, library_id=None, history_id=None, workflow=None, workflow_inputs=None, workflow_params=None, dbkey='?', method=None, folder=None, link_data_only=None):
    if library_id:
        file_info['upload_json'] = _upload_to_library(gi, file_info, method, library_id, folder_id=folder['id'], dbkey=dbkey, link_data_only=link_data_only)
        src = 'ldda'
    elif history_id:
        file_info['upload_json'] = gi.tools.upload_file(file_info['file_location'], history_id, file_type=file_info['file_type'], dbkey=None)
        src = 'hda'
    if workflow:
        if workflow_inputs:
            inputs = workflow_inputs
        else:
            inputs = []
        inputs.append("{0}={1}={2}".format(file_info['label'], file_info['upload_json'][0]['id'], src))
    if 'add_files' in file_info:
        for f in file_info['add_files']:
            _do_upload(gi, f, library_id=library_id, history_id=history_id, dbkey=dbkey, method=method, folder=folder, link_data_only=link_data_only)
            if workflow:
                inputs.append("{0}={1}={2}".format(f['label'], f['upload_json'][0]['id'], src))
    if workflow:
        inputs = _parse_inputs(inputs, workflow)
        if not history_id:
            history = {'name': "{0} - {1} - {2}".format(folder['name'], file_info['upload_json'][0]['name'], workflow['name'])}
            history = _get_history(gi, history)
            history_id = history['id']
        workflow['run_json'] = gi.workflows.run_workflow(workflow['id'], inputs, params=workflow_params, history_id=history_id)
        pp = pprint.PrettyPrinter(indent=4)
        pp.pprint(workflow['run_json'])

def _parse_params(param_list):
    params = {}
    for p in param_list:
        tool, param, value = p.split('=')
        if tool is None or param is None or value is None:
            raise ValueError("Wrong parameter syntax: {0}.".format(p))
        params[tool] = {'param': param, 'value': value}
    return params

def _parse_inputs(input_list, workflow):
    inputs = {}
    label_map = {}
    for i in input_list:
        input_id, dataset_id, src = i.split('=')
        if input_id is None or dataset_id is None or src is None:
            raise ValueError("Wrong input syntax: {0}.".format(i))
        inputs[input_id] = {'id': dataset_id, 'src': src}
    for i in workflow['inputs']:
        if i not in inputs:
            if workflow['inputs'][i]['label'] in inputs:
                inputs[i] = inputs[workflow['inputs'][i]['label']]
            elif '.' in inputs:
                inputs[i] = inputs['.']
            else:
                raise ValueError("No input was provided for workflow input: 'id: {0}, label: {1}'.".format(i, workflow['inputs'][i]['label']))
    return inputs

if __name__ == '__main__':
    exit()

